package index;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.util.ArrayList;
import java.util.TreeSet;

import javax.swing.JComboBox;
import javax.swing.JFrame;
import javax.swing.JScrollBar;
import javax.swing.JTextField;
import javax.swing.ListCellRenderer;
import javax.swing.plaf.basic.BasicComboPopup;

/**
 * A TextField which supports autocompletion. It also supports completion
 * of multiple values sepparated with a, well, sepparator.
 * More infos @ http://klaue.110mb.com/
 * @author Klaue
 *
 */
public class AutoCompletionTextField extends JTextField {
	private static final long serialVersionUID = -5226333779181474237L;
	JComboBox possibleValues = new JComboBox();
	BasicComboPopup menu = new BasicComboPopup(this.possibleValues);
	private CustomActionListener cmboActList = new CustomActionListener();
	Character cSepparator = null;
	private ArrayList<String> alWords = new ArrayList<String>();
	private final int MAX_VISIBLE_ROW_COUNT = 4;
	private final int WIDTH_OF_SCROLLBAR = new JScrollBar().getPreferredSize().width;
	
	/**
	 * Constructs a new AutoCompletionTextField.
	 * @param sepparator The Character to sepparate the words/sentences which
	 * 					 should be autocompleted. To not use sepparators, set it to null.
	 * @param words The ArrayList which contains the words/sentences which should be
	 * 				displayed in the drop-down-list
	 */
	public AutoCompletionTextField(Character sepparator, ArrayList<String> words){
		this(words);
		setSepparator(sepparator);
	}
	
	/**
	 * Constructs a new AutoCompletionTextField. The initial String is null,
	 * the numbers of columns is set to 0 and the sepparator is set to null
	 * @param words The ArrayList which contains the words/sentences which should be
	 * 				displayed in the drop-down-list
	 */
	public AutoCompletionTextField(ArrayList<String> words) {
		super();
		setWords(words);
		addKeyListener(new KeyAdapter(){
			@Override
			public void keyReleased(KeyEvent e) {
				redrawSuggestionList(e);
			}
		});
	}
	
	/**
	 * Constructs a new AutoCompletionTextField. The initial String is set to text and
	 * the numbers of columns is set to 0.
	 * @param text The text to be displayed or null if no text should be displayed
	 * @param sepparator The Character to sepparate the words/sentences which
	 * 					 should be autocompleted. To not use sepparators, set it to null.
	 * @param words The ArrayList which contains the words/sentences which should be
	 * 				displayed in the drop-down-list
	 */
	public AutoCompletionTextField(String text, Character sepparator, ArrayList<String> words){
		this(text, words);
		setSepparator(sepparator);
	}
	
	/**
	 * Constructs a new AutoCompletionTextField. The initial String is set to text,
	 * the numbers of columns is set to 0 and the sepparator is set to null.
	 * @param text The text to be displayed or null if no text should be displayed
	 * @param words The ArrayList which contains the words/sentences which should be
	 * 				displayed in the drop-down-list
	 */
	public AutoCompletionTextField(String text, ArrayList<String> words) {
		super(text);
		setWords(words);
		addKeyListener(new KeyAdapter(){
			@Override
			public void keyReleased(KeyEvent e) {
				redrawSuggestionList(e);
			}
		});
	}
	
	/**
	 * Sets the sepparator. Null for no sepparator
	 * @param c the sepparator
	 */
	public void setSepparator(Character c) {
		this.cSepparator = c;
	}
	
	/**
	 * Sets the list of words which should be used for completion
	 * @param words The list
	 */
	public void setWords(ArrayList<String> words) {
		this.alWords = words;
	}
	
	/**
	 * @return The sepparator or null for no sepparator
	 */
	public Character getSepparator() {
		return this.cSepparator;
	}
	
	/**
	 * @return The list of words used for Autocompletion
	 */
	public ArrayList<String> getWords() {
		return this.alWords;
	}
	
	/**
	 * Checks if the menu should be displayed and if so, display it
	 * @param keyEvent the pressed key
	 */
	void redrawSuggestionList(KeyEvent keyEvent){
		if (this.alWords.size() == 0) return;
		int keyCode = keyEvent.getKeyCode();
		requestFocus();
		
		// exclude left/right keys
		if (keyCode == 37 || keyCode == 39) return;
		
		// process up/down-keys
		if (keyCode == 38 || keyCode == 40) {
			if (this.possibleValues.getItemCount() == 1) return;
			int i = this.possibleValues.getSelectedIndex();
			if (i == -1) return;
			
			// remove actionlistener, so that the menu doesn't go away when the selection is changed
			this.possibleValues.removeActionListener(this.cmboActList);
			if (keyCode == 38) { // up
				if (i == 0) {
					this.possibleValues.setSelectedIndex(this.possibleValues.getItemCount() - 1);
				} else {
					this.possibleValues.setSelectedIndex(i - 1);
				}
			} else { // down
				if (i == this.possibleValues.getItemCount() - 1) {
					this.possibleValues.setSelectedIndex(0);
				} else {
					this.possibleValues.setSelectedIndex(i + 1);
				}
			}
			// re-add the action listener
			this.possibleValues.addActionListener(this.cmboActList);
			return;
		}
		
		// enter
		if (keyCode == 10) {
			if (this.possibleValues.getSelectedIndex() == -1) return;
			// fire actionevent
			this.possibleValues.setSelectedIndex(this.possibleValues.getSelectedIndex());
			return;
		}
		
		//menu.setVisible(false);
		this.menu.removeAll();
		
		String query = getText();
		
		if (query == null || query.trim().length() == 0 ||
				(this.cSepparator != null && query.trim().endsWith("" + this.cSepparator))) {
			this.menu.hide();
			return;
		}
		
		// get new query
		if (this.cSepparator != null) {
			int start = query.lastIndexOf("" + this.cSepparator, getCaretPosition() - 1);
			int end = query.indexOf(this.cSepparator, getCaretPosition());
			if (end == -1) end = query.length();
			query = query.substring(start + 1, end).trim();
			if (query.trim().length() == 0) {
				this.menu.hide();
				return;
			}
		}

		TreeSet<String> items = new TreeSet<String>();
		for (String s : this.alWords) {
			if (s.toLowerCase().startsWith(query.toLowerCase())) {
				items.add(s);
			}
		}
		if (items.size() == 0) {
			this.menu.setVisible(false);
			return;
		}
		
		this.possibleValues = new JComboBox(items.toArray());
		this.possibleValues.addActionListener(this.cmboActList);
		this.menu = new BasicComboPopup(this.possibleValues);
		
		this.menu.setPreferredSize(getPopupSize(this.MAX_VISIBLE_ROW_COUNT));
		
		int xLocation = getLocation().x;
		if (this.cSepparator != null) {
			int i = getText().lastIndexOf(this.cSepparator + "", getCaretPosition() - 1);
			if (i != -1) {
				//String s = getText().substring(0, i + 1);
				String s = getText().substring(i + 1, getCaretPosition());
				int countOfWhitespace = s.length() - ((s + "*").trim().length() - 1); // "*" = trim does not affect right side
				s = getText().substring(0, i + 1 + countOfWhitespace);
				int lengthOfString = this.getGraphics().getFontMetrics(this.getFont()).stringWidth(s) + 2;  // + 2 = "border" of the textfield
				xLocation += lengthOfString - getHorizontalVisibility().getValue();
			}
		}
		this.menu.show(this, xLocation, getHeight());
	}
	
	/**
	 * Imitates the protected method getPopupHeightForRowCount in BasicComboPopup
	 * with added functionality for the width
	 * @param maxRowCount Number of rows which should be displayed at once
	 * @return The dimension for the menu
	 */
	private Dimension getPopupSize(int maxRowCount) {
		int totalWidth = this.menu.getPreferredSize().width;
		int totalHeight = 0;
		ListCellRenderer rend = this.menu.getList().getCellRenderer();

		if (this.possibleValues.getItemCount() > maxRowCount) {
			totalWidth += this.WIDTH_OF_SCROLLBAR;
		}
		
		if (this.possibleValues.getItemCount() < maxRowCount) {
			maxRowCount = this.possibleValues.getItemCount();
		}

		for (int i = 0; i < maxRowCount; i++) {
			Component comp = rend.getListCellRendererComponent(this.menu.getList(),
							this.possibleValues.getModel().getElementAt(i), -1,
							false, false);
			Dimension dim = comp.getPreferredSize();
			totalHeight += dim.height;
		}

		// plus 2 because else the scrollbar will be displayed
		return new Dimension(totalWidth, totalHeight + 2);
		//return totalHeight == 0 ? 100 : totalHeight + 2;
	}
	
	/**
	 * Sets the text in the appropriate place of this to the selected item of possibleValues 
	 * @author Klaue
	 */
	class CustomActionListener implements ActionListener{
		public void actionPerformed(ActionEvent e) {
			if (!AutoCompletionTextField.this.menu.isVisible()) return;
			String text = (String)AutoCompletionTextField.this.possibleValues.getSelectedItem();
			int caretPos = text.length();
			if (AutoCompletionTextField.this.cSepparator != null) {
				int start = getText().lastIndexOf("" + AutoCompletionTextField.this.cSepparator, getCaretPosition() - 1);
				int end = getText().indexOf(AutoCompletionTextField.this.cSepparator, getCaretPosition());
				if (end == -1) end = getText().length();
				if (start != -1) {
					text = getText().substring(0, start + 1) + " " + text;
				}
				caretPos = text.length();
				text += getText().substring(end);
			}
			setText(text);
			setCaretPosition(caretPos);
			AutoCompletionTextField.this.menu.hide();
		}
	}
 
	public static void main(String[] args) {
		
//		 Create and set up the window.
		JFrame frame = new JFrame("SuggestList");
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
 
		// Create and set up the content pane.
		ArrayList<String> artists = new ArrayList<String>();
		artists.add("Nek0gami");
		artists.add("Venator Somniorum");
		artists.add("Norbert");
		artists.add("Nekalderig");
		frame.add(new AutoCompletionTextField(' ', artists));
		//newContentPane.setOpaque(true); // content panes must be opaque
		//frame.setContentPane(newContentPane);
 
		// Display the window.
		frame.pack();
		frame.setVisible(true);
	}
 
}

